@@ -52,10 +52,13 @@ struct GsmXSMPClientPrivate
         guint      watch_id;
         guint      protocol_timeout;
 
-        int        current_save_yourself;
-        int        next_save_yourself;
         char      *description;
         GPtrArray *props;
+
+        /* SaveYourself state */
+        int        current_save_yourself;
+        int        next_save_yourself;
+        guint      next_save_yourself_allow_interact : 1;
 };
 
 enum {
@@ -221,6 +224,7 @@ gsm_xsmp_client_init (GsmXSMPClient *client)
         client->priv->props = g_ptr_array_new ();
         client->priv->current_save_yourself = -1;
         client->priv->next_save_yourself = -1;
+        client->priv->next_save_yourself_allow_interact = FALSE;
 }
 
 
@@ -403,7 +407,6 @@ xsmp_get_restart_command (GsmClient *client)
         return prop_to_command (prop);
 }
 
-#if 0
 static char *
 xsmp_get_discard_command (GsmClient *client)
 {
@@ -417,12 +420,11 @@ xsmp_get_discard_command (GsmClient *client)
 
         return prop_to_command (prop);
 }
-#endif
 
 static void
 do_save_yourself (GsmXSMPClient *client,
                   int            save_type,
-                  gboolean       forceful)
+                  gboolean       allow_interact)
 {
         g_assert (client->priv->conn != NULL);
 
@@ -437,8 +439,12 @@ do_save_yourself (GsmXSMPClient *client,
                 g_debug ("GsmXSMPClient:   queuing new SaveYourself for '%s'",
                          client->priv->description);
                 client->priv->next_save_yourself = save_type;
+                client->priv->next_save_yourself_allow_interact = allow_interact;
         } else {
                 client->priv->current_save_yourself = save_type;
+                /* make sure we don't have anything queued */
+                client->priv->next_save_yourself = -1;
+                client->priv->next_save_yourself_allow_interact = FALSE;
 
                 switch (save_type) {
                 case SmSaveLocal:
@@ -452,7 +458,7 @@ do_save_yourself (GsmXSMPClient *client,
 
                 default:
                         /* Logout */
-                        if (forceful) {
+                        if (!allow_interact) {
                                 SmsSaveYourself (client->priv->conn,
                                                  save_type, /* save type */
                                                  TRUE, /* shutdown */
@@ -470,7 +476,6 @@ do_save_yourself (GsmXSMPClient *client,
         }
 }
 
-#if 0
 static void
 xsmp_save_yourself_phase2 (GsmClient *client)
 {
@@ -490,7 +495,6 @@ xsmp_interact (GsmClient *client)
 
         SmsInteract (xsmp->priv->conn);
 }
-#endif
 
 static gboolean
 xsmp_cancel_end_session (GsmClient *client,
@@ -509,6 +513,12 @@ xsmp_cancel_end_session (GsmClient *client,
         }
 
         SmsShutdownCancelled (xsmp->priv->conn);
+
+        /* reset the state */
+        xsmp->priv->current_save_yourself = -1;
+        xsmp->priv->next_save_yourself = -1;
+        xsmp->priv->next_save_yourself_allow_interact = FALSE;
+
         return TRUE;
 }
 
@@ -620,9 +630,10 @@ static GKeyFile *
 xsmp_save (GsmClient *client,
            GError   **error)
 {
-        GKeyFile *keyfile;
+        GKeyFile *keyfile = NULL;
         char     *desktop_file_path = NULL;
         char     *exec_program = NULL;
+        char     *exec_discard = NULL;
         char     *startup_id = NULL;
         GError   *local_error;
 
@@ -631,6 +642,11 @@ xsmp_save (GsmClient *client,
 
         local_error = NULL;
 
+        exec_program = xsmp_get_restart_command (client);
+        if (!exec_program) {
+                goto out;
+        }
+
         desktop_file_path = get_desktop_file_path (GSM_XSMP_CLIENT (client));
 
         keyfile = create_client_key_file (client,
@@ -650,15 +666,22 @@ xsmp_save (GsmClient *client,
                                "X-GNOME-Autostart-startup-id",
                                startup_id);
 
-        exec_program = xsmp_get_restart_command (client);
         g_key_file_set_string (keyfile,
                                G_KEY_FILE_DESKTOP_GROUP,
                                G_KEY_FILE_DESKTOP_KEY_EXEC,
                                exec_program);
 
+        exec_discard = xsmp_get_discard_command (client);
+        if (exec_discard)
+                g_key_file_set_string (keyfile,
+                                       G_KEY_FILE_DESKTOP_GROUP,
+                                       "X-GNOME-Autostart-discard-exec",
+                                       exec_discard);
+
 out:
         g_free (desktop_file_path);
         g_free (exec_program);
+        g_free (exec_discard);
         g_free (startup_id);
 
         if (local_error != NULL) {
@@ -689,7 +712,8 @@ xsmp_query_end_session (GsmClient *client,
                         guint      flags,
                         GError   **error)
 {
-        gboolean forceful;
+        gboolean allow_interact;
+        int      save_type;
 
         if (GSM_XSMP_CLIENT (client)->priv->conn == NULL) {
                 g_set_error (error,
@@ -699,8 +723,16 @@ xsmp_query_end_session (GsmClient *client,
                 return FALSE;
         }
 
-        forceful = (flags & GSM_CLIENT_END_SESSION_FLAG_FORCEFUL);
-        do_save_yourself (GSM_XSMP_CLIENT (client), SmSaveGlobal, forceful);
+        allow_interact = !(flags & GSM_CLIENT_END_SESSION_FLAG_FORCEFUL);
+
+        /* we don't want to save the session state, but we just want to know if
+         * there's user data the client has to save and we want to give the
+         * client a chance to tell the user about it. This is consistent with
+         * the manager not setting GSM_CLIENT_END_SESSION_FLAG_SAVE for this
+         * phase. */
+        save_type = SmSaveGlobal;
+
+        do_save_yourself (GSM_XSMP_CLIENT (client), save_type, allow_interact);
         return TRUE;
 }
 
@@ -709,7 +741,7 @@ xsmp_end_session (GsmClient *client,
                   guint      flags,
                   GError   **error)
 {
-        gboolean forceful;
+        gboolean phase2;
 
         if (GSM_XSMP_CLIENT (client)->priv->conn == NULL) {
                 g_set_error (error,
@@ -719,8 +751,28 @@ xsmp_end_session (GsmClient *client,
                 return FALSE;
         }
 
-        forceful = (flags & GSM_CLIENT_END_SESSION_FLAG_FORCEFUL);
-        do_save_yourself (GSM_XSMP_CLIENT (client), SmSaveGlobal, forceful);
+        phase2 = (flags & GSM_CLIENT_END_SESSION_FLAG_LAST);
+
+        if (phase2) {
+                xsmp_save_yourself_phase2 (client);
+        } else {
+                gboolean allow_interact;
+                int      save_type;
+
+                /* we gave a chance to interact to the app during
+                 * xsmp_query_end_session(), now it's too late to interact */
+                allow_interact = FALSE;
+
+                if (flags & GSM_CLIENT_END_SESSION_FLAG_SAVE) {
+                        save_type = SmSaveBoth;
+                } else {
+                        save_type = SmSaveGlobal;
+                }
+
+                do_save_yourself (GSM_XSMP_CLIENT (client),
+                                  save_type, allow_interact);
+        }
+
         return TRUE;
 }
 
@@ -1099,7 +1151,7 @@ save_yourself_request_callback (SmsConn   conn,
                 g_signal_emit (client, signals[LOGOUT_REQUEST], 0, !fast);
         } else if (!shutdown && !global) {
                 g_debug ("GsmXSMPClient:   initiating checkpoint");
-                do_save_yourself (client, SmSaveLocal, FALSE);
+                do_save_yourself (client, SmSaveLocal, TRUE);
         } else {
                 g_debug ("GsmXSMPClient:   ignoring");
         }
@@ -1114,21 +1166,13 @@ save_yourself_phase2_request_callback (SmsConn   conn,
         g_debug ("GsmXSMPClient: Client '%s' received SaveYourselfPhase2Request",
                  client->priv->description);
 
-        /* Treat this just like a SaveYourselfDone */
-        if (client->priv->current_save_yourself == SmSaveLocal) {
-                client->priv->current_save_yourself = -1;
-                SmsSaveComplete (client->priv->conn);
-        } else {
-                client->priv->current_save_yourself = -1;
-        }
+        client->priv->current_save_yourself = -1;
 
         /* this is a valid response to SaveYourself and therefore
            may be a response to a QES or ES */
-        gdm_client_end_session_response (GSM_CLIENT (client),
-                                         TRUE,
+        gsm_client_end_session_response (GSM_CLIENT (client),
+                                         TRUE, TRUE, FALSE,
                                          NULL);
-
-        /* FIXME: change priority so this client runs at the end? */
 }
 
 static void
@@ -1137,17 +1181,20 @@ interact_request_callback (SmsConn   conn,
                            int       dialog_type)
 {
         GsmXSMPClient *client = manager_data;
+#if 0
         gboolean       res;
         GError        *error;
+#endif
 
         g_debug ("GsmXSMPClient: Client '%s' received InteractRequest(%s)",
                  client->priv->description,
                  dialog_type == SmDialogNormal ? "Dialog" : "Errors");
 
-        gdm_client_end_session_response (GSM_CLIENT (client),
-                                         FALSE,
+        gsm_client_end_session_response (GSM_CLIENT (client),
+                                         FALSE, FALSE, FALSE,
                                          _("This program is blocking log out."));
 
+#if 0
         /* Can't just call back with Interact because session client
            grabs the keyboard!  So, we try to get it to release
            grabs by telling it we've cancelled the shutdown.
@@ -1159,6 +1206,8 @@ interact_request_callback (SmsConn   conn,
                 g_warning ("Unable to cancel end session: %s", error->message);
                 g_error_free (error);
         }
+#endif
+        xsmp_interact (GSM_CLIENT (client));
 }
 
 static void
@@ -1172,8 +1221,8 @@ interact_done_callback (SmsConn   conn,
                  client->priv->description,
                  cancel_shutdown ? "True" : "False");
 
-        gdm_client_end_session_response (GSM_CLIENT (client),
-                                         TRUE,
+        gsm_client_end_session_response (GSM_CLIENT (client),
+                                         TRUE, FALSE, cancel_shutdown,
                                          NULL);
 }
 
@@ -1188,28 +1237,25 @@ save_yourself_done_callback (SmsConn   conn,
                  client->priv->description,
                  success ? "True" : "False");
 
-        if (client->priv->current_save_yourself == SmSaveLocal) {
-                client->priv->current_save_yourself = -1;
-                SmsSaveComplete (client->priv->conn);
-        } else {
-                client->priv->current_save_yourself = -1;
-        }
+	if (client->priv->current_save_yourself != -1) {
+		SmsSaveComplete (client->priv->conn);
+		client->priv->current_save_yourself = -1;
+	}
 
-        /* If success is false then the application still has
-           unsafe data.  We may also have tricked it into sending
-           us this message when we faked the ShutdownCancel to
-           break its grabs. */
-        if (success) {
-                gdm_client_end_session_response (GSM_CLIENT (client),
-                                                 TRUE,
-                                                 NULL);
-        }
+        /* If success is false then the application couldn't save data. Nothing
+         * the session manager can do about, though. FIXME: we could display a
+         * dialog about this, I guess. */
+        gsm_client_end_session_response (GSM_CLIENT (client),
+                                         TRUE, FALSE, FALSE,
+                                         NULL);
 
         if (client->priv->next_save_yourself) {
-                int save_type = client->priv->next_save_yourself;
+                int      save_type = client->priv->next_save_yourself;
+                gboolean allow_interact = client->priv->next_save_yourself_allow_interact;
 
                 client->priv->next_save_yourself = -1;
-                do_save_yourself (client, save_type, FALSE);
+                client->priv->next_save_yourself_allow_interact = -1;
+                do_save_yourself (client, save_type, allow_interact);
         }
 }
 
